@@ -7,11 +7,13 @@ INI-style "tox.ini" file.
 from __future__ import print_function
 
 import os
+import pipes
 import re
 import shutil
 import subprocess
 import sys
 import time
+from contextlib import contextmanager
 
 import pkg_resources
 import py
@@ -19,6 +21,7 @@ import py
 import tox
 from tox.config import parseconfig
 from tox.result import ResultLog
+from tox.util import set_os_env_var
 from tox.venv import VirtualEnv
 
 
@@ -42,28 +45,35 @@ def cmdline(args=None):
 def main(args):
     try:
         config = prepare(args)
-        retcode = Session(config).runcommand()
+        with set_os_env_var("TOX_WORK_DIR", config.toxworkdir):
+            retcode = build_session(config).runcommand()
         if retcode is None:
             retcode = 0
         raise SystemExit(retcode)
     except KeyboardInterrupt:
         raise SystemExit(2)
-    except tox.exception.MinVersionError as e:
+    except (tox.exception.MinVersionError, tox.exception.MissingRequirement) as e:
         r = Reporter(None)
         r.error(str(e))
         raise SystemExit(1)
 
 
+def build_session(config):
+    return Session(config)
+
+
 def show_help(config):
     tw = py.io.TerminalWriter()
     tw.write(config._parser._format_help())
     tw.line()
     tw.line("Environment variables", bold=True)
     tw.line("TOXENV: comma separated list of environments (overridable by '-e')")
+    tw.line("TOX_SKIP_ENV: regular expression to filter down from running tox environments")
     tw.line(
         "TOX_TESTENV_PASSENV: space-separated list of extra environment variables to be "
         "passed into test command environments"
     )
+    tw.line("PY_COLORS: 0 disable colorized output, 1 enable (default)")
 
 
 def show_help_ini(config):
@@ -72,7 +82,7 @@ def show_help_ini(config):
     for env_attr in config._testenv_attr:
         tw.line(
             "{:<15} {:<8} default: {}".format(
-                env_attr.name, "<" + env_attr.type + ">", env_attr.default
+                env_attr.name, "<{}>".format(env_attr.type), env_attr.default
             ),
             bold=True,
         )
@@ -136,9 +146,16 @@ class Action(object):
     def popen(self, args, cwd=None, env=None, redirect=True, returnout=False, ignore_ret=False):
         stdout = outpath = None
         resultjson = self.session.config.option.resultjson
+
+        cmd_args = [str(x) for x in args]
+        cmd_args_shell = " ".join(pipes.quote(i) for i in cmd_args)
         if resultjson or redirect:
             fout = self._initlogpath(self.id)
-            fout.write("actionid: {}\nmsg: {}\ncmdargs: {!r}\n\n".format(self.id, self.msg, args))
+            fout.write(
+                "actionid: {}\nmsg: {}\ncmdargs: {!r}\n\n".format(
+                    self.id, self.msg, cmd_args_shell
+                )
+            )
             fout.flush()
             outpath = py.path.local(fout.name)
             fin = outpath.open("rb")
@@ -153,16 +170,18 @@ class Action(object):
             popen = self._popen(args, cwd, env=env, stdout=stdout, stderr=subprocess.STDOUT)
         except OSError as e:
             self.report.error(
-                "invocation failed (errno {:d}), args: {}, cwd: {}".format(e.errno, args, cwd)
+                "invocation failed (errno {:d}), args: {}, cwd: {}".format(
+                    e.errno, cmd_args_shell, cwd
+                )
             )
             raise
         popen.outpath = outpath
-        popen.args = [str(x) for x in args]
+        popen.args = cmd_args
         popen.cwd = cwd
         popen.action = self
         self._popenlist.append(popen)
         try:
-            self.report.logpopen(popen, env=env)
+            self.report.logpopen(popen, cmd_args_shell)
             try:
                 if resultjson and not redirect:
                     if popen.stderr is not None:
@@ -274,13 +293,12 @@ class Reporter(object):
         else:
             return Verbosity.DEBUG
 
-    def logpopen(self, popen, env):
+    def logpopen(self, popen, cmd_args_shell):
         """ log information about the action.popen() created process. """
-        cmd = " ".join(map(str, popen.args))
         if popen.outpath:
-            self.verbosity1("  {}$ {} >{}".format(popen.cwd, cmd, popen.outpath))
+            self.verbosity1("  {}$ {} >{}".format(popen.cwd, cmd_args_shell, popen.outpath))
         else:
-            self.verbosity1("  {}$ {} ".format(popen.cwd, cmd))
+            self.verbosity1("  {}$ {} ".format(popen.cwd, cmd_args_shell))
 
     def logaction_start(self, action):
         msg = "{} {}".format(action.msg, " ".join(map(str, action.args)))
@@ -300,13 +318,14 @@ class Reporter(object):
         if self.verbosity >= Verbosity.QUIET:
             self.tw.sep("_", "summary")
 
-    def info(self, msg):
-        if self.verbosity >= Verbosity.DEBUG:
-            self.logline(msg)
+    def logline_if(self, level, msg, key=None, **kwargs):
+        if self.verbosity >= level:
+            message = str(msg) if key is None else "{}{}".format(key, msg)
+            self.logline(message, **kwargs)
 
-    def using(self, msg):
-        if self.verbosity >= Verbosity.INFO:
-            self.logline("using {}".format(msg), bold=True)
+    def logline(self, msg, **opts):
+        self.reported_lines.append(msg)
+        self.tw.line("{}".format(msg), **opts)
 
     def keyboard_interrupt(self):
         self.error("KEYBOARDINTERRUPT")
@@ -321,40 +340,32 @@ class Reporter(object):
     def line(self, msg, **opts):
         self.logline(msg, **opts)
 
+    def info(self, msg):
+        self.logline_if(Verbosity.DEBUG, msg)
+
+    def using(self, msg):
+        self.logline_if(Verbosity.INFO, msg, "using ", bold=True)
+
     def good(self, msg):
-        if self.verbosity >= Verbosity.QUIET:
-            self.logline(msg, green=True)
+        self.logline_if(Verbosity.QUIET, msg, green=True)
 
     def warning(self, msg):
-        if self.verbosity >= Verbosity.QUIET:
-            self.logline("WARNING: {}".format(msg), red=True)
+        self.logline_if(Verbosity.QUIET, msg, "WARNING: ", red=True)
 
     def error(self, msg):
-        if self.verbosity >= Verbosity.QUIET:
-            self.logline("ERROR: {}".format(msg), red=True)
+        self.logline_if(Verbosity.QUIET, msg, "ERROR: ", red=True)
 
     def skip(self, msg):
-        if self.verbosity >= Verbosity.QUIET:
-            self.logline("SKIPPED: {}".format(msg), yellow=True)
-
-    def logline(self, msg, **opts):
-        self.reported_lines.append(msg)
-        self.tw.line("{}".format(msg), **opts)
+        self.logline_if(Verbosity.QUIET, msg, "SKIPPED: ", yellow=True)
 
     def verbosity0(self, msg, **opts):
-        if self.verbosity >= Verbosity.DEFAULT:
-            self.logline("{}".format(msg), **opts)
+        self.logline_if(Verbosity.DEFAULT, msg, **opts)
 
     def verbosity1(self, msg, **opts):
-        if self.verbosity >= Verbosity.INFO:
-            self.logline("{}".format(msg), **opts)
+        self.logline_if(Verbosity.INFO, msg, **opts)
 
     def verbosity2(self, msg, **opts):
-        if self.verbosity >= Verbosity.DEBUG:
-            self.logline("{}".format(msg), **opts)
-
-    # def log(self, msg):
-    #    print(msg, file=sys.stderr)
+        self.logline_if(Verbosity.DEBUG, msg, **opts)
 
 
 class Session:
@@ -371,7 +382,7 @@ class Session:
         self._spec2pkg = {}
         self._name2venv = {}
         try:
-            self.venvlist = [self.getvenv(x) for x in self.config.envlist]
+            self.venvlist = [self.getvenv(x) for x in self.evaluated_env_list()]
         except LookupError:
             raise SystemExit(1)
         except tox.exception.ConfigError as e:
@@ -379,6 +390,18 @@ class Session:
             raise SystemExit(1)
         self._actions = []
 
+    def evaluated_env_list(self):
+        tox_env_filter = os.environ.get("TOX_SKIP_ENV")
+        tox_env_filter_re = re.compile(tox_env_filter) if tox_env_filter is not None else None
+        for name in self.config.envlist:
+            if tox_env_filter_re is not None and tox_env_filter_re.match(name):
+                msg = "skip environment {}, matches filter {!r}".format(
+                    name, tox_env_filter_re.pattern
+                )
+                self.report.verbosity1(msg)
+                continue
+            yield name
+
     @property
     def hook(self):
         return self.config.pluginmanager.hook
@@ -419,7 +442,24 @@ class Session:
         elif self.config.option.listenvs_all:
             self.showenvs(all_envs=True, description=verbosity)
         else:
-            return self.subcommand_test()
+            with self.cleanup():
+                return self.subcommand_test()
+
+    @contextmanager
+    def cleanup(self):
+        self.config.temp_dir.ensure(dir=True)
+        try:
+            yield
+        finally:
+            for tox_env in self.venvlist:
+                if (
+                    hasattr(tox_env, "package")
+                    and isinstance(tox_env.package, py.path.local)
+                    and tox_env.package.exists()
+                ):
+                    self.report.verbosity2("cleanup {}".format(tox_env.package))
+                    tox_env.package.remove()
+                    py.path.local(tox_env.package.dirname).remove(ignore_errors=True)
 
     def _copyfiles(self, srcdir, pathlist, destdir):
         for relpath in pathlist:
@@ -470,7 +510,7 @@ class Session:
                 )
             except tox.exception.InterpreterNotFound as e:
                 status = e
-                if self.config.option.skip_missing_interpreters:
+                if self.config.option.skip_missing_interpreters == "true":
                     default_ret_code = 0
             if status:
                 str_status = str(status)
@@ -496,8 +536,8 @@ class Session:
             try:
                 venv.developpkg(setupdir, action)
                 return True
-            except tox.exception.InvocationError:
-                venv.status = sys.exc_info()[1]
+            except tox.exception.InvocationError as exception:
+                venv.status = exception
                 return False
 
     def installpkg(self, venv, path):
@@ -513,8 +553,8 @@ class Session:
             try:
                 venv.installpkg(path, action)
                 return True
-            except tox.exception.InvocationError:
-                venv.status = sys.exc_info()[1]
+            except tox.exception.InvocationError as exception:
+                venv.status = exception
                 return False
 
     def subcommand_test(self):
@@ -522,9 +562,10 @@ class Session:
             self.report.info("skipping sdist step")
         else:
             for venv in self.venvlist:
-                venv.package = self.hook.tox_package(session=self, venv=venv)
-                if not venv.package:
-                    return 2
+                if not venv.envconfig.skip_install:
+                    venv.package = self.hook.tox_package(session=self, venv=venv)
+                    if not venv.package:
+                        return 2
         if self.config.option.sdistonly:
             return
         for venv in self.venvlist:
@@ -556,14 +597,15 @@ class Session:
         envlog.set_installed(packages)
 
     def runtestenv(self, venv, redirect=False):
-        if not self.config.option.notest:
+        if self.config.option.notest:
+            venv.status = "skipped tests"
+        else:
             if venv.status:
                 return
             self.hook.tox_runtest_pre(venv=venv)
-            self.hook.tox_runtest(venv=venv, redirect=redirect)
+            if venv.status == 0:
+                self.hook.tox_runtest(venv=venv, redirect=redirect)
             self.hook.tox_runtest_post(venv=venv)
-        else:
-            venv.status = "skipped tests"
 
     def _summary(self):
         self.report.startsummary()
@@ -572,7 +614,7 @@ class Session:
             status = venv.status
             if isinstance(status, tox.exception.InterpreterNotFound):
                 msg = " {}: {}".format(venv.envconfig.envname, str(status))
-                if self.config.option.skip_missing_interpreters:
+                if self.config.option.skip_missing_interpreters == "true":
                     self.report.skip(msg)
                 else:
                     retcode = 1
@@ -620,7 +662,7 @@ class Session:
         env_conf = self.config.envconfigs  # this contains all environments
         default = self.config.envlist  # this only the defaults
         ignore = {self.config.isolated_build_env}.union(default)
-        extra = sorted(e for e in env_conf if e not in ignore) if all_envs else []
+        extra = [e for e in env_conf if e not in ignore] if all_envs else []
 
         if description:
             self.report.line("default environments:")
